@@ -22,8 +22,9 @@ from .nmeasource import (
     ModemError,
     NmeaSourceNotFoundError,
     QuectelNmeaSource,
+    GnssShareNmeaSource,
 )
-from .util import now, unique_filename, bearing_to_arrow
+from .util import bearing_to_arrow, have_touchscreen, now, unique_filename
 from .widgets import text_barchart, DataFrame
 from satellite import __version__
 
@@ -32,12 +33,15 @@ gi.require_version('Gdk', '3.0')
 gi.require_version('Handy', '1')
 from gi.repository import Gdk, Gio, GLib, Gtk, Handy  # noqa: E402
 
+appname = 'Satellite'
+app_id = 'page.codeberg.tpikonen.satellite'
+
 
 class SatelliteApp(Gtk.Application):
 
     def __init__(self, *args, **kwargs):
         Gtk.Application.__init__(
-            self, *args, application_id="page.codeberg.tpikonen.satellite",
+            self, *args, application_id=app_id,
             flags=Gio.ApplicationFlags.FLAGS_NONE, **kwargs)
         Handy.init()
 
@@ -47,6 +51,12 @@ class SatelliteApp(Gtk.Application):
             '-c', '--console-output', dest='console_output',
             action='store_true', default=False,
             help='Output satellite data to console')
+        parser.add_argument(
+            '-s', '--source', dest='source',
+            default='quectel',
+            help='Select NMEA source. Options are '
+            '\'quectel\' (default) for Quectel Modems or '
+            '\'gnss-share\' when using gnss-share')
         self.args = parser.parse_args()
 
         GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, signal.SIGINT,
@@ -71,7 +81,6 @@ class SatelliteApp(Gtk.Application):
                         self, widget_api_name))
             else:
                 setattr(self, widget_api_name, widget)
-        self.connect("activate", self.do_activate)
 
         self.app_menu = self.builder.get_object('app-menu')
         self.menu_popover = Gtk.Popover.new_from_model(
@@ -99,10 +108,13 @@ class SatelliteApp(Gtk.Application):
         self.set_speedlabel(None)
         self.leaflet.set_visible_child(self.databox)
 
-        self.datascroll.connect('edge-overshot', self.on_edge_overshot)
+        self.connect('startup', self.on_startup)
+        self.connect('activate', self.on_activate)
+        self.connect('shutdown', self.on_shutdown)
 
         # Internal state
         self.last_mode = 1
+        self.last_data = None
         self.last_speed = None
         self.last_update = None
         self.source_lost = False
@@ -147,46 +159,40 @@ class SatelliteApp(Gtk.Application):
             Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
         )
 
-    def do_startup(self):
-        Gtk.Application.do_startup(self)
+    def on_startup(self, app):
         self.create_actions()
         # Initialize modem after GUI startup
         GLib.idle_add(self.init_source)
 
-    def do_activate(self, *args):
+    def on_activate(self, app):
         self.setup_styles()
         self.add_window(self.window)
         self.window.show()
-        return True
+        if have_touchscreen():
+            self.datascroll.connect('edge-overshot', self.on_edge_overshot)
+
+    def on_shutdown(self, app):
+        """Called after main loop exits."""
+        print("Cleaning up...")
+        self.gpx_write()
+        if self.source is not None:
+            self.source.close()
+        print("...done.")
 
     def init_source(self):
-        self.log_msg(f"Satellite version {__version__} started")
-        try:
-            self.source = QuectelNmeaSource(
-                self.location_update_cb,
-                refresh_rate=self.refresh_rate,
-                # save_filename=unique_filename(self.gpx_save_dir + '/nmeas',
-                #                               '.txt')
-            )
-            self.source.initialize()
-        except Exception as e:
-            fatal = False
-            if isinstance(e, ModemLockedError):
-                self.log_msg("Modem is locked")
-                dtext = "Please unlock the Modem"
-            else:
-                self.log_msg("Error initializing NMEA source")
-                dtext = e.message if hasattr(e, 'message') else (
-                    "Could not find or initialize NMEA source")
-            dialog = Gtk.MessageDialog(
-                parent=self.window, modal=True,
-                message_type=Gtk.MessageType.ERROR,
-                buttons=Gtk.ButtonsType.OK, text=dtext)
-            dialog.set_title("Error initializing NMEA source")
-            dialog.run()
-            dialog.destroy()
-            if fatal:
-                self.quit()
+        self.log_msg(f"{appname} version {__version__} started")
+        source_init = False
+
+        self.log_msg(f'Trying to initialize source "{self.args.source}"')
+
+        if self.args.source == 'quectel':
+            source_init = self.init_quectel_source()
+        elif self.args.source == 'gnss-share':
+            source_init = self.init_gnss_share_source()
+
+        if not source_init:
+            self.log_msg('No NmeaSource initialized')
+            return False  # Remove from idle_add
 
         self.log_msg(
             f"Source is {self.source.manufacturer}, model {self.source.model}"
@@ -210,15 +216,65 @@ class SatelliteApp(Gtk.Application):
 
         GLib.timeout_add(self.refresh_rate * 1000, self.timeout_cb, None)
 
-        return False  # Remove from idle_add
+        return GLib.SOURCE_REMOVE
 
-    def quit_function(self):
-        """Called after main loop exits."""
-        print("Cleaning up...")
-        self.gpx_write()
-        if self.source is not None:
-            self.source.restore()
-        print("...done.")
+    def init_gnss_share_source(self):
+        try:
+            self.source = GnssShareNmeaSource(self.location_update_cb)
+            self.source.initialize()
+        except Exception as e:
+            fatal = False
+            self.log_msg(str(e))
+            dtext = 'Can you access "/var/run/gnss-share.sock"?\n' + str(e)
+            dialog = Gtk.MessageDialog(
+                parent=self.window, modal=True,
+                message_type=Gtk.MessageType.ERROR,
+                buttons=Gtk.ButtonsType.OK, text=dtext)
+            dialog.set_title("Error initializing NMEA source")
+            dialog.run()
+            dialog.destroy()
+            if fatal:
+                self.quit()
+
+            return False
+
+        return True
+
+    def init_quectel_source(self):
+        try:
+            self.source = QuectelNmeaSource(
+                self.location_update_cb,
+                refresh_rate=self.refresh_rate,
+                # save_filename=unique_filename(self.gpx_save_dir + '/nmeas',
+                #                               '.txt')
+            )
+            self.source.initialize()
+        except Exception as e:
+            fatal = False
+            if isinstance(e, ModemLockedError):
+                self.log_msg("Modem is locked")
+                dtext = "Please unlock the Modem"
+            else:
+                fatal = isinstance(e, gi.repository.GLib.GError)
+                self.log_msg("Error initializing NMEA source")
+                etext = str(e)
+                dtext = etext if etext else (
+                    "Could not find or initialize NMEA source")
+            dialog = Gtk.MessageDialog(
+                parent=self.window, modal=True,
+                message_type=Gtk.MessageType.ERROR,
+                buttons=Gtk.ButtonsType.CLOSE if fatal else Gtk.ButtonsType.OK,
+                text=dtext)
+            dialog.set_title("Error initializing NMEA source")
+            dialog.run()
+            dialog.destroy()
+            if fatal:
+                self.quit()
+                return
+
+            return False
+
+        return True
 
     def sigint_handler(self):
         if not self.sigint_received:
@@ -240,8 +296,8 @@ class SatelliteApp(Gtk.Application):
         adlg = Gtk.AboutDialog(
             transient_for=self.window,
             modal=True,
-            program_name="Satellite",
-            logo_icon_name="page.codeberg.tpikonen.satellite",
+            program_name=appname,
+            logo_icon_name=app_id,
             version=__version__,
             comments="A program for showing navigation satellite data",
             license_type=Gtk.License.GPL_3_0_ONLY,
@@ -311,31 +367,31 @@ class SatelliteApp(Gtk.Application):
 
     def leaflet_forward_cb(self, button):
         self.leaflet.navigate(Handy.NavigationDirection.FORWARD)
-        return True
 
     def leaflet_back_cb(self, button):
         self.leaflet.navigate(Handy.NavigationDirection.BACK)
-        return True
 
     def infolabel_released_cb(self, gesture, n_press, x, y):
         if n_press != 1 or self.carousel.get_position() > 0.5:
-            return False
+            return
         self.chart_size = self.chart_small if (
             self.chart_size == self.chart_large) else self.chart_large
-        self.update(None)
-        return True
+        self.set_barchart(self.last_data)
 
     def carousel_page_changed_cb(self, carousel, index):
         if index == 1 and self.chart_size == self.chart_large:
             self.chart_size = self.chart_small
-            self.update(None)
+            self.set_barchart(self.last_data)
         return False
 
-    def format_satellite_data(self, d):
-        bchart = text_barchart(
-            ((e['prn'], e['snr']) for e in d['visibles']),
-            d['actives'], height=self.chart_size)
-        return bchart
+    def set_barchart(self, data):
+        if data is None:
+            return ''
+        barchart = text_barchart(
+            ((e['prn'], e['snr']) for e in data['visibles']),
+            data['actives'], height=self.chart_size)
+        self.infolabel.set_markup("<tt>" + barchart + "</tt>")
+        return barchart
 
     def set_values(self, data):
         def to_str(x, fmt="%s"):
@@ -471,19 +527,22 @@ class SatelliteApp(Gtk.Application):
     def timeout_cb(self, x):
         dt = (time.time() - self.last_update) if self.last_update else 100
         if dt > 2 * self.refresh_rate:
-            self.update(None)
-        return True
+            self.main_box.set_sensitive(False)
+            self.update()
+        return GLib.SOURCE_CONTINUE
 
     def location_update_cb(self, *args):
         self.last_update = time.time()
-        self.update(None)
+        self.main_box.set_sensitive(True)
+        self.update()
 
-    def update(self, x):
+    def update(self):
         try:
             nmeas = self.source.get()
         except Exception as e:
             fatal = False
             show_dialog = False
+            etext = str(e)
             dtext = None
             if isinstance(e, ModemLockedError):
                 dtext = "Please unlock the Modem"
@@ -491,15 +550,14 @@ class SatelliteApp(Gtk.Application):
             elif isinstance(e, ModemNoNMEAError):
                 dtext = "NMEA info not received with location"
             elif isinstance(e, ModemError):
-                dtext = "Unspecified modem error"
+                dtext = "Modem error: " + str(e)
             elif isinstance(e, NmeaSourceNotFoundError):
                 if not self.source_lost:
-                    dtext = e.message if (
-                        hasattr(e, 'message')) else "Modem disappeared"
+                    dtext = etext if etext else "Modem disappeared"
                     self.source_lost = True
                     self.main_box.set_sensitive(False)
             else:
-                dtext = e.message if hasattr(e, 'message') else "Unknown error"
+                dtext = etext if etext else "Unknown error"
             if show_dialog:
                 dialog = Gtk.MessageDialog(
                     parent=self.window, modal=True,
@@ -510,25 +568,30 @@ class SatelliteApp(Gtk.Application):
                 dialog.destroy()
                 if fatal:
                     self.quit()
+                    return
             elif dtext is not None:
                 self.log_msg(dtext)
-            return True
+            return
+
+        if not nmeas:
+            return
 
         if self.source_lost:
             self.log_msg("Modem appeared")
             self.main_box.set_sensitive(True)
 
         self.source_lost = False
+
         data = nmea.parse(nmeas)
         data["updateage"] = ((time.time() - self.last_update)
                              if self.last_update else None)
-        barchart = self.format_satellite_data(data)
+
+        barchart = self.set_barchart(data)
         if self.args.console_output:
             print(barchart)
-        self.infolabel.set_markup("<tt>" + barchart + "</tt>")
+
         self.set_values(data)
-        mode = data["mode"]
-        mode = int(mode) if mode else 0
+
         speed = data['speed']
         bearing = data['true_course']
         self.set_speedlabel(speed, bearing)
@@ -537,13 +600,18 @@ class SatelliteApp(Gtk.Application):
         elif not speed and self.last_speed:
             self.carousel.scroll_to(self.infolabel)
         self.last_speed = speed
+
         # log
+        mode = data["mode"]
+        mode = int(mode) if mode else 0
         if mode != self.last_mode:
             if mode > 1:
                 self.log_msg(f"Got lock, mode: {mode}")
             elif mode <= 1:
                 self.log_msg("Lock lost")
         self.last_mode = mode
+
         if self.gpx is not None and data.get("valid"):
             self.gpx_update(data)
-        return True
+
+        self.last_data = data
